"""

Test config construction

Simple tests that ensure that the Config construction isn't broken

"""

import unittest
import logging

import configerus
from configerus.config import Config

logger = logging.getLogger("basic_construct")


class BasicConstruct(unittest.TestCase):

    def test_construct_1_naked(self):
        """ just make sure we can create a Configs object, without any bootstraps """
        config = configerus.new_config(bootstraps=[])
        self.assertIsInstance(config, Config)

    def test_construct_2_single_bootstrap(self):
        """ make sure we can create a Configs object, without a single bootstrap """
        configerus.new_config(bootstraps=['dict'])

    def test_construct_3_default_bootsraps(self):
        """ Make sure we can create a Configs object, with default bootstraps """
        configerus.new_config()

    def test_construct_3_bad_bootsraps(self):
        """ Make sure we can create a Configs object, with default bootstraps """
        with self.assertRaises(KeyError):
            configerus.new_config(bootstraps=['I do not exist'])



from typing import List, Dict, Any

from configerus.config import Config


class ConfigFormatDeepPlugin():
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

    def format(self, target, default_source: str):
        """ Deep formatter for formatting that reruns formatting across iterables

        this formatter just reruns formatting on elements of primitive iterables.

        """
        # try to iterate across the target
        # There is probably a more `python` approach that would cover more
        # iterables, as long as we can re-assign
        if isinstance(target, Dict):
            return self.format_dict(target, default_source)
        elif isinstance(target, List):
            return self.format_list(target, default_source)

        # anything we can't iterate through we just return
        return target

    def format_list(self, target: List[Any], default_source: str):
        """ Search a List for strings to format @see format_string """
        for index, value in enumerate(target):
            target[index] = self.config.format(value, default_source)
        return target

    def format_dict(self, target: Dict[str, Any], default_source: str):
        """ Search a Dict for strings to format @see format_string """
        for key, value in target.items():
            target[key] = self.config.format(value, default_source)
        return target



import logging

from configerus.config import Config
from configerus.plugin import SourceFactory

from .env_specific_source import ConfigSourceEnvSpecificPlugin, PLUGIN_ID_SOURCE_ENV_SPECIFIC, CONFIGERUS_ENV_SPECIFIC_BASE_KEY
from .env_json_source import ConfigSourceEnvJsonPlugin, PLUGIN_ID_SOURCE_ENV_JSON, CONFIGERUS_ENV_JSON_ENV_KEY


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_ENV_SPECIFIC)
def plugin_factory_configsource_env(config: Config, instance_id: str = ''):
    """ create an configsource specific env plugin """
    return ConfigSourceEnvSpecificPlugin(config, instance_id)


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_ENV_JSON)
def plugin_factory_configsource_env(config: Config, instance_id: str = ''):
    """ create an configsource json env plugin """
    return ConfigSourceEnvJsonPlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object

    We don't actually do anything, so this bootstrapper is here only to ensure
    that the above factory decorators are run

    """
    pass


# export the imported constants to make it easier for consumers to import
__all__ = [
    plugin_factory_configsource_env,
    plugin_factory_configsource_env,
    configerus_bootstrap,
    PLUGIN_ID_SOURCE_ENV_SPECIFIC,
    CONFIGERUS_ENV_SPECIFIC_BASE_KEY,
    PLUGIN_ID_SOURCE_ENV_JSON,
    CONFIGERUS_ENV_JSON_ENV_KEY]



"""

test the plugin instance functionality

"""
import logging
import unittest

from configerus.config import Config
from configerus.plugin import SourceFactory, FormatFactory, ValidatorFactory, Type
from configerus.instances import PluginInstance, PluginInstances

logger = logging.getLogger("test_instances")

""" register a bunch of dummy plugins which do nothing but can be used for testing """


@SourceFactory(plugin_id='dummy_1')
def plugin_factory_source_1(config: Config, instance_id: str = ''):
    """ create a dummy configsource plugin """
    return DummyPlugin(config, instance_id)


@SourceFactory(plugin_id='dummy_2')
def plugin_factory_source_2(config: Config, instance_id: str = ''):
    """ create a dummy configsource plugin """
    return DummyPlugin(config, instance_id)


@FormatFactory(plugin_id='dummy_1')
def plugin_factory_format_1(config: Config, instance_id: str = ''):
    """ create a dummy formatter plugin """
    return DummyPlugin(config, instance_id)


@FormatFactory(plugin_id='dummy_2')
def plugin_factory_format_2(config: Config, instance_id: str = ''):
    """ create a dummy formatter plugin """
    return DummyPlugin(config, instance_id)


@ValidatorFactory(plugin_id='dummy_1')
def plugin_factory_validate_1(config: Config, instance_id: str = ''):
    """ create a dummy formatter plugin """
    return DummyPlugin(config, instance_id)


@ValidatorFactory(plugin_id='dummy_2')
def plugin_factory_validate_2(config: Config, instance_id: str = ''):
    """ create a dummy formatter plugin """
    return DummyPlugin(config, instance_id)


class DummyPlugin:
    """ Just a placehold plugin object  """

    def __init__(self, config: Config, instance_id: str):
        self.config = config
        self.instance_id = instance_id


class ConfigTemplating(unittest.TestCase):

    def test_instance_struct_sanity(self):
        """ just test that the Instance struct has the properties that we use """
        config = Config()

        instance = PluginInstance(
            plugin_id='dummy',
            instance_id='dummy',
            type=Type.SOURCE,
            priority=60,
            plugin=DummyPlugin(config, 'dummy')
        )

        self.assertEqual(instance.plugin_id, 'dummy')
        self.assertEqual(instance.priority, 60)

    def test_instances_sanity(self):
        """ Test some construction sanity on the instances object """
        config = Config()
        instances = PluginInstances(config.make_plugin)

        self.assertEqual(len(instances), 0)

        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_instance_1', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_instance_2', config.default_priority())

        self.assertEqual(len(instances), 2)

    def test_instances_plugin_get_simple(self):
        """ test that we can add plugins and then retrieve them """
        config = Config()
        instances = PluginInstances(config.make_plugin)

        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_instance_1', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_instance_2', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_2', 'dummy_instance_3', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_2', 'dummy_instance_4', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_instance_5', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_2', 'dummy_instance_6', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_instance_7', config.default_priority())

        len(instances) == 7

        get_4 = instances['dummy_instance_4']
        self.assertTrue(get_4)
        self.assertEqual(get_4.instance_id, 'dummy_instance_4')
        self.assertEqual(instances.get_plugin(instance_id='dummy_instance_4').instance_id, 'dummy_instance_4')
        with self.assertRaises(KeyError):
            instances.get_plugin(instance_id='no_such_instance')

    def test_instances_get_plugins(self):
        """ get plugins based on multiple search parameters """
        config = Config()
        instances = PluginInstances(config.make_plugin)

        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_source_1', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_1', 'dummy_source_2', config.default_priority())
        instances.add_plugin(Type.SOURCE, 'dummy_2', 'dummy_instance', config.default_priority())
        instances.add_plugin(Type.FORMATTER, 'dummy_2', 'dummy_formatter_4', config.default_priority())
        instances.add_plugin(Type.FORMATTER, 'dummy_1', 'dummy_formatter_5', config.default_priority())
        instances.add_plugin(Type.FORMATTER, 'dummy_2', 'dummy_formatter_6', config.default_priority())
        instances.add_plugin(Type.VALIDATOR, 'dummy_1', 'dummy_instance', config.default_priority())
        instances.add_plugin(Type.VALIDATOR, 'dummy_2', 'dummy_validator_8', config.default_priority())

        self.assertTrue(instances.has_plugin(instance_id='dummy_formatter_4'))
        self.assertTrue(instances.has_plugin(instance_id='dummy_validator_8'))
        self.assertTrue(instances.has_plugin(instance_id='dummy_formatter_4'))
        self.assertTrue(instances.has_plugin(instance_id='dummy_source_2'))

        self.assertEqual(len(instances.get_instances(type=Type.SOURCE)), 3)
        self.assertEqual(len(instances.get_instances(type=Type.FORMATTER)), 3)
        self.assertEqual(len(instances.get_instances(type=Type.VALIDATOR)), 2)

        self.assertEqual(len(instances.get_instances(type=Type.SOURCE, plugin_id='dummy_1')), 2)
        self.assertEqual(len(instances.get_instances(type=Type.VALIDATOR, plugin_id='dummy_1')), 1)

        # Plugin ID only search should cross type
        self.assertEqual(len(instances.get_instances(plugin_id='dummy_1')), 4)
        # There are no rules against repeated instance_ids
        self.assertEqual(len(instances.get_instances(instance_id='dummy_instance')), 2)
        # no rule against empty filtering (gets sorted instances)
        self.assertEqual(len(instances.get_instances()), len(instances))

    def test_instance_priority_simple(self):
        """ test that retrieving instances sorts """
        config = Config()
        instances = PluginInstances(config.make_plugin)

        starting_range = range(30, 71, 10)
        priority_list = range(70, 29, -10)

        for priority in starting_range:
            instances.add_plugin(Type.SOURCE, 'dummy_1', 'instance_{}'.format(priority), priority)

        instance_list = instances.get_instances(type=Type.SOURCE)
        self.assertEqual(len(instance_list), len(starting_range))

        for index, priority in enumerate(priority_list):
            self.assertEqual(instance_list[index].priority, priority)

        # let's see what happens as things change
        instances.add_plugin(Type.SOURCE, 'dummy_1', 'instance_{}'.format(90), 90)
        instance_list = instances.get_instances(type=Type.SOURCE)

        self.assertEqual(len(instance_list), len(starting_range) + 1)
        self.assertEqual(instance_list[0].priority, 90)
        self.assertEqual(instance_list[1].priority, 70)
        self.assertEqual(instance_list[len(starting_range)].priority, 30)



from typing import Dict, Any
import logging

from configerus.config import Config


class ConfigSourceDictPlugin():
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.data = {}
        """ keep the data that we will use for searching """

    def set_data(self, data: Dict[str, Any]):
        """ Assign Dict data to this config source plugin """
        self.data = data

    def load(self, label: str):
        """ Load a config label and return a Dict[str, Any] of config data

        Parameters:

        label (str) : label to load

        """
        if label in self.data:
            return self.data[label]
        else:
            return {}




from configerus.config import Config
from configerus.plugin import FormatFactory, SourceFactory

from .source import ConfigSourcePathPlugin
from .format import ConfigFormatFilePlugin


CONFIGERUS_PATH_KEY = 'path'
""" if you load paths from config, this is the key that should give the path """

PLUGIN_ID_SOURCE_PATH = 'path'
""" ConfigSource plugin_id for the configerus path configsource plugin """


@SourceFactory(plugin_id=PLUGIN_ID_SOURCE_PATH)
def plugin_factory_configsource_path(config: Config, instance_id: str = ''):
    """ create an configsource path plugin """
    return ConfigSourcePathPlugin(config, instance_id)


PLUGIN_ID_FORMAT_FILE = 'configerus.plugin.formatter.file'
""" Format plugin_id for the configerus filepath format plugin """


@FormatFactory(plugin_id=PLUGIN_ID_FORMAT_FILE)
def plugin_factory_format_file(config: Config, instance_id: str = ''):
    """ create an format plugin which replaces from file contents """
    return ConfigFormatFilePlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object """
    config.add_formatter(plugin_id=PLUGIN_ID_FORMAT_FILE, instance_id=PLUGIN_ID_FORMAT_FILE, priority=40)



import logging
from tempfile import mkdtemp
from shutil import rmtree
from typing import List
import os.path
import json
import yaml

from configerus.config import Config
from configerus.contrib.env import PLUGIN_ID_SOURCE_ENV_SPECIFIC
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH

logger = logging.getLogger('configerus:tests')

TEST_CONFIG_TEMP_DIR = ''
""" on first use, we will populate this with a str path to a parent tmp dir that
    will contain any file config that we use """


def make_test_config(config: Config, sources: List):
    """ Make config of various types from a list, and add it to a config object """
    global TEST_CONFIG_TEMP_DIR

    logger.info("Building sources from source list")
    for source in sources:
        name = source["name"]
        priority = source["priority"] if "priority" in source else config.default_priority()
        type = source["type"]
        data = source["data"]

        if type == PLUGIN_ID_SOURCE_DICT:
            logger.info("Adding 'dict' source '%s' [%s]: %s", name, priority, data.keys())
            config.add_source(type, name, priority).set_data(data)

        elif type == PLUGIN_ID_SOURCE_PATH:
            # on first use, get a temp dir
            if not TEST_CONFIG_TEMP_DIR:
                TEST_CONFIG_TEMP_DIR = mkdtemp()

            # first make files for all of the data
            path = name
            full_path = os.path.join(TEST_CONFIG_TEMP_DIR, path)
            os.makedirs(full_path)
            for file_name, file_data in data.items():
                full_file = os.path.join(full_path, file_name)
                logger.debug("path source '%s' writing file '%s' : %s", name, full_file, file_data)
                with open(full_file, 'w') as config_file_pointer:
                    extension = os.path.splitext(file_name)[1].lower()[1:]
                    if extension == "json":
                        json.dump(file_data, config_file_pointer)
                    elif extension == "yaml" or extension == "yml":
                        yaml.dump(file_data, config_file_pointer)

            # then add the created paths to the config sources
            logger.info("Adding 'path' source '%s' [%s]: %s : %s", name, priority, path, data.keys())
            config.add_source(type, name, priority).set_path(full_path)

        if type == PLUGIN_ID_SOURCE_ENV_SPECIFIC:
            # we can do env plugins as well
            logger.info("Adding 'env' source '%s' [%s]: %s", name, priority, data.keys())
            for env_key, env_data in data.items():
                env_name = "{}_{}".format('CONFIGERUSTEST', env_key).upper()
                logger.info("Adding environment variable : {}".format(env_name))
                os.environ[env_name] = str(env_data)
            config.add_source(type, name, priority).set_base('CONFIGERUSTEST')


def test_config_cleanup(config: Config):
    """ clean up any created temporary folder created """
    if TEST_CONFIG_TEMP_DIR:
        rmtree(TEST_CONFIG_TEMP_DIR)




import os.path
import re
import yaml
import json
import logging

from configerus.config import Config

FILES_FORMAT_MATCH_PATTERN = r'(\[(file\:)(?P<file>(\~?\/?\w+\/)*\w*(\.\w+)?)\])'
""" A regex pattern to identify files that should be embedded """

logger = logging.getLogger('configerus.contrib.files:format')


class ConfigFormatFilePlugin:
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.pattern = re.compile(FILES_FORMAT_MATCH_PATTERN)

    def format(self, target, default_source: str):
        """ Format a string by substituting config values

        Parameters
        ----------

        target: a string that should be formatted. If not a string then no
            formatting is performed

        default_source : if format/replace patterns don't have a source defined
            then this is used as a source.

        """
        if not isinstance(target, str):
            return target

        # if the entire target is the match, then replace whatever type we get
        # out of the config .get() call
        match = self.pattern.fullmatch(target)
        if match:
            return self._get(match, return_only_string=False)

        # search through the target replacing any found matches with
        start = 0
        match = self.pattern.search(target, start)
        while match:
            rep = str(self._get(match))
            target = target[:match.start()] + rep + target[match.end()]
            start = start + len(rep)
            match = self.pattern.search(target, start)

        return target

    def _get(self, match, return_only_string: bool = True):
        """ find a file match and return the file contents

        Parameters
        ----------

        match (re Match) a regex match which we use to determine file path

        return_only_string (bool) optional indictor that the function is
            expecting only a string return.  Knowing this means that we can
            skip unmarshalling file contents which could be expensive.

        Raises
        ------

        FileNotFound is a file replacement is requested but the file path
            cannot be opened

        json.decoder.JSONDecodeError if a json file cannot be unmarshalled

        yaml.YAMLError if a yaml file cannot be unmarshalled

        Returns
        -------

        unmarshalled json/yml file or string contents of the file

        """
        file = match.group('file')
        extension = ''

        file_split = os.path.splitext(file)
        if len(file_split) > 0:
            extension = file_split[1].lower()

        try:
            with open(file) as file_o:
                if not return_only_string:
                    # try to parse/unmarshall a file instead of just returing it

                    if extension == ".json":
                        try:
                            data = json.load(file_o)
                        except json.decoder.JSONDecodeError as e:
                            raise ValueError(
                                "Failed to parse one of the config files '{}': {}".format(
                                    os.path.join(
                                        self.path, file), e))
                        return data

                    elif extension == ".yml" or extension == ".yaml":
                        try:
                            data = yaml.load(file_o, Loader=yaml.FullLoader)
                        except yaml.YAMLError as e:
                            raise ValueError(
                                "Failed to parse one of the config files '{}': {}".format(
                                    os.path.join(
                                        self.path, file), e))
                        return data

                # return file contents as a string (above parsing didn't happen)
                return file_o.read()

        except FileNotFoundError as e:
            raise KeyError("Could not embed file as config as file could not be found: {}".format(file))



"""

Config source from JSON contents of an ENV variable

This plugin allows you to include JSON in a single ENV variable which if present
will be used as a config source.

This allows large runtime config overrides with out a lot of individual ENV
variables being needed.

If you find that your JSON is tiny and difficult to design/manage, then confider
usng the 'specific' env source plugin.

string case for keys is a weird topic.  We .lower() all keys as it seems a
decent convention.

"""
from typing import Dict, Any
import logging
import os

from configerus.config import Config

logger = logging.getLogger('configerus.contrib.env.source.specific')

PLUGIN_ID_SOURCE_ENV_SPECIFIC = 'env-specific'
""" ConfigSource plugin_id for the configerus env configsource plugin """

CONFIGERUS_ENV_SPECIFIC_BASE_KEY = 'base'
""" Config key for retreiving the env specific base value from config """


class ConfigSourceEnvSpecificPlugin():
    """ Get config from all ENV variables """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.base = ''
        """ keep the ENV base prefix that should limit what gets pulled in """

    def set_base(self, base: str):
        self.base = base

    def load(self, label: str):
        """ Load a config label and return a Dict[str, Any] of config data

        Parameters:

        label (str) : label to load

        """

        if self.base:
            label_prefix = '{}_{}_'.format(self.base, label).upper()
        else:
            label_prefix = '{}_'.format(label).upper()

        filtered = {}
        for env_key, env_value in os.environ.items():
            if env_key.upper().startswith(label_prefix):
                filtered[env_key[len(label_prefix):].lower()] = env_value

        organized = {}
        for (key, value) in filtered.items():
            if key.startswith('_'):
                continue

            base = organized
            key_steps = key.lower().split('_')

            last_step = key_steps.pop()

            for key_step in key_steps:
                if not isinstance(base, dict):
                    base = {}
                if not key_step in base:
                    base[key_step] = {}
                base = base[key_step]
            if not isinstance(base, dict):
                base = {}
            base[last_step] = value

        return organized



import logging
from typing import Dict, Any, List

logger = logging.getLogger('configerus.shared')

# @see https://stackoverflow.com/questions/20656135/python-deep-merge-dictionary-data


def tree_merge(source: Any, destination: Any):
    """
    Deep merge source into destination

    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
    True
    """

    if not (isinstance(source, dict) and isinstance(destination, dict)):
        return source

    for key, value in source.items():
        if isinstance(value, dict) and key in destination:
            value = tree_merge(value, destination[key])
        destination[key] = value

    return destination


def tree_get(node: Dict, keys: List[str], glue: str = '.', ignore: List[str] = []) -> Any:
    """ Find a path down a tree using the keys as a step by step path """

    if not node:
        raise ValueError("There was no data in the config so no key match could be made")

    flat_steps = tree_reduce(tree=keys, glue=glue, ignore=ignore)
    if len(flat_steps) == 0:
        return node

    for step in flat_steps:
        try:
            if node is None:
                raise KeyError("Path tried to descend into None")
            if isinstance(node, str):
                raise KeyError("Path tried to descend into a string: {}".format(node))
            elif isinstance(node, list) and step.isnumeric():
                node = node[int(step)]
            else:
                # hopefully the target is subscriptable?
                node = node[step]

        except KeyError as e:
            raise KeyError("Key {} not found in loaded config data. '{}' was not found".format(keys, step)) from e
        except IndexError as e:
            raise IndexError("Array index '{}' was not found in list : {}".format(step, node))
        except TypeError as e:
            raise ValueError("Invalid key '{}' in the keys list: '{}' : {}".format(step, node, e)) from e

    return node


def tree_reduce(tree: Any, glue: str = '.', ignore: List[Any] = []) -> List[str]:
    """ merge a nested tree of strings down to a flat list of strings

    Also split any strings that contain a glue character into a list of strings
    and flatten them down as well.

    Also ignore certain passed in strings from the list.

    Parameters:
    -----------

    tree (str | List[str] | List[List[str]] ...) : Any list tree of strings that
        need to be flattened to a single depth

    glue (str) : if a string is found that has a glue, then it is split along
        that glue and converted to a List

    ignore (List[str]) : a list of strings that should be considered invalid
        values that should not be included in the flattenned list

    Returns:

    List[str] tree reduced to a flat (single) depth, potentially empty

    """

    # don't use the iterator solution with strings
    if isinstance(tree, str):
        if glue == '':
            tree = [tree]
        else:
            tree = tree.split(glue)

    else:
        flatter = []
        for node in tree:
            flat_node = tree_reduce(node, glue, ignore)
            flatter += flat_node
        tree = flatter

    return [node for node in tree if node and node not in ignore]




from configerus.config import Config
from configerus.plugin import ValidatorFactory

from .validate import JsonSchemaValidatorPlugin

PLUGIN_ID_VALIDATE_JSONSCHEMA = 'jsonschema'
""" Format plugin_id for the configerus jsonschema validator plugin """
PLUGIN_PRIORITY_VALIDATE_JSONSCHEMA_PRIORITY = 60
""" bootstrapping jsonschema validate plugin priority """


@ValidatorFactory(plugin_id=PLUGIN_ID_VALIDATE_JSONSCHEMA)
def plugin_factory_validator_schema(config: Config, instance_id: str = ''):
    """ create a validate plugin which applies json schema validation """
    return JsonSchemaValidatorPlugin(config, instance_id)


def configerus_bootstrap(config: Config):
    """ Bootstrap a config object by adding our formatter """
    config.add_validator(PLUGIN_ID_VALIDATE_JSONSCHEMA, priority=PLUGIN_PRIORITY_VALIDATE_JSONSCHEMA_PRIORITY)



import logging
from typing import Any
from .shared import tree_get, tree_reduce

logger = logging.getLogger('configerus:loaded')

LOADED_KEY_ROOT = ''
""" If this key is requested in .get() then the root data dict is returned """


class Loaded:
    """ A loaded config which contains all of the file config for a single label

    This is an easy to useconfig object for a single load. From this you can get
    config using dot notation.

    Think of this as relating to a single filename, merged from different paths
    as opposed to the Config object which loads config and hands off to this one.

    """

    def __init__(self, data, parent, instance_id: str):
        """
        parameters
        ----------

        data (Dict[str, Any]): deep dict struct that contains all of the merged
           configuration that is te be used for config retrieval. It is often
           a nested Dict with standard primitives as can be loaded from json/yml

        parent (Config): The parent Config object which created this object
           which is used for backreferencing, primarily when trying to perform
           template string substitution as substitution can refer to config from
           other sources.

        """
        assert data is not None, "None data was passed in"

        self.data = data
        self.parent = parent
        self.instance_id = instance_id

    def _reload(self, data):
        """ Force new data to be used

        This is meant to be used by the parent only, and is used to update config
        when new config paths are added, typically by boostrapping.

        Hopeully this is done before the config is really used.

        @TODO signal on this?

        """
        self.data = data

    def has(self, key: Any = LOADED_KEY_ROOT, format: bool = True,
            exception_if_missing: bool = False, validator: str = ""):
        """ check if a key value exists in the config

        using a key, traverse a tree of data and return a boolean for if the key
        can be found.

        Parameters:

        key (Varies) : specify the key path down the loaded config data dict.

            This is a very forgiving and adaptive parameter.  You have can pass:

            '1.2.3.4' : "dot" notation which means the steps down the tree are
                1, 2, 3 then 4.

            ['1', '2', '3', '4'] : asks for the same '1.2.3.4' path as above.

            if the loaded data was:
            ```
                {
                    '1': {
                        '2': {
                            '3': {
                                '4': 'my value'
                            }
                        }
                    }
                }
            ```
            Then the returned value would be 'my value'

            You can also mix and match with out worrying about clean formatting:

            ['1.2', '3.4']
            ['1.2.3.4', '', '.']
            ['1', ['2', '3.4']]
            ['1', '2', '3.', '.4']

            (all are equivalent to '1.2.3.4')

            This allows you to pass around keys and append to them on the fly
            without having to be overly concerned about clean formatting.

        Returns:
        --------

        Boolean : if a value exists in laoded config

        """
        return self.get(key, format=False, validator='', exception_if_missing=False) is not None

    def get(self, key: Any = LOADED_KEY_ROOT, format: bool = True,
            exception_if_missing: bool = False, validator: str = ""):
        """ get a key value from the loaded config

        using a key, traverse a tree of data and return the value, optionally
        formatted, optionally validated.

        Parameters:

        key (Varies) : specify the key path down the loaded config data dict.

            This is a very forgiving and adaptive parameter.  You have can pass:

            '1.2.3.4' : "dot" notation which means the steps down the tree are
                1, 2, 3 then 4.

            ['1', '2', '3', '4'] : asks for the same '1.2.3.4' path as above.

            if the loaded data was:
            ```
                {
                    '1': {
                        '2': {
                            '3': {
                                '4': 'my value'
                            }
                        }
                    }
                }
            ```
            Then the returned value would be 'my value'

            You can also mix and match with out worrying about clean formatting:

            ['1.2', '3.4']
            ['1.2.3.4', '', '.']
            ['1', ['2', '3.4']]
            ['1', '2', '3.', '.4']

            (all are equivalent to '1.2.3.4')

            this allows you to pass around keys and append to them on the fly
            without having to be overly concerned about clean formatting.

        format (bool): should retrieved string values be checked for variable
           substitution?  If so then the str value is checked using regex for
           things that should be replaced with other config values.
           @see self.format_string().)

           Formatting IS NOT APPLIED if the key could not be matched.

        validator (str) : string key passed to the validate() method which will
            validate the retrieved data before returning it.

            Validation IS APPLIED if the key could not be matched.

        Returns:

        (Any) anything in the Dict is a valid return.  The return could be a
            Dict with child elements, an array or any other primitive.
            If the return is a string then it is formatted for variable
            substitution (see self.format_string().)

        Throws:

        Can throw a KeyError if the key cannot be found (which also occurs if
        all sources produced no data and a non-empty key was passed.)

        """
        value = ""

        try:
            value = tree_get(self.data, key, ignore=['', LOADED_KEY_ROOT])

        except KeyError as e:
            if exception_if_missing:
                # hand off the exception
                raise e
            else:
                logger.debug("Failed to find config key : %s", key)
                value = None

        if format and value is not None:
            # try to format any values
            value = self.format(value)

        if validator:
            self.parent.validate(value, validator)

        return value

    def format(self, data):
        """ Format some data using the config object formatters

        Parameters:

        data (Any): primitive data that should be formatted. The data will be
            passed to the formatter plugins in descending priority order.

        """
        return self.parent.format(data, self.instance_id)

    def validate(self, data, validate_target: Any, exception_if_invalid: bool = True):
        """ Validate some data using the config object validators

        data (Any): primitive data that should be validated. The data will be
            passed to the validator plugins in descending priority order.

        validate_target (Any) : Validation target which will be interpreted by the
            validation plugins.  Two typical formats are common:

            1. a string with some identifiable pattern that a plugi can recognize
            2. a dict with keys that different plugins can recognize.

            if empty/None then no validation is performed.

        Returns:
        --------

        Bool results of validation. If any validation plugin raises an exception
        then it will be false.

        If exception_if_invalid then validation failures with raise and exception

        Raises:
        -------

        ValidationError on any validation exception.

        ValdiationError if any validator ran into an exception internally.

        @TODO can we be more specific in interpreting validation exceptions.

        """
        return self.parent.validate(data=data, validate_target=validate_target,
                                    exception_if_invalid=exception_if_invalid)



"""

Get config overrides for specific config values from different ENV variables

With this plugin, all ENV variables are interpreted as different config values
dependening on their name, and the values are added as a config source.

This allows small single ENV variable overrides for specific values at runtime.

If you find that you are using a large number of ENV variables for your
overrides, consider using the JSON plugin instead.

"""
from typing import Dict, Any
import logging
import os
import json

from configerus.config import Config

logger = logging.getLogger('configerus.contrib.env.source.json')

PLUGIN_ID_SOURCE_ENV_JSON = 'env-json'
""" ConfigSource plugin_id for the configerus json env configsource plugin """

CONFIGERUS_ENV_JSON_ENV_KEY = 'env'
""" Config key for retreiving the env json env value from config """


class ConfigSourceEnvJsonPlugin():
    """ Get config from  a single ENV variables """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.source = {}
        """ default to empty source """

    def set_env(self, env: str):
        logger.debug("Setting new ENV source: {}".format(env))
        env_json = os.getenv(env)

        # Allow the env variable to be empty
        if env_json is None or env_json == '':
            self.source = {}
        else:
            try:
                self.source = json.loads(env_json)
            except json.decoder.JSONDecodeError as e:
                raise ValueError("Invalid json in {} ENV variable: {}".format(env, e)) from e

    def load(self, label: str):
        """ Load a config label and return a Dict[str, Any] of config data

        Parameters:

        label (str) : label to load

        """
        if label in self.source:
            return self.source[label]

        return {}
